! The subroutines of the main program
!**************************************************************

subroutine Gram_Schmidt_orthogonormalization(n,NNN,znorm,y)
implicit none
integer,intent(in)		:: n, NNN
real(kind=8),intent(inout)	:: znorm(n), y(NNN)
real(kind=8)			:: gsc(n)
integer				:: j, k, l

znorm(1)=0.d0
do j=1,n
	znorm(1)=znorm(1)+y(n*j+1)**2
enddo
znorm(1)=dsqrt(znorm(1))
 
do j=1,n
	y(n*j+1)=y(n*j+1)/znorm(1)
enddo
do j=2,n
	do k=1,j-1
		gsc(k)=0.d0
		do l=1,n
			gsc(k)=gsc(k)+y(n*l+j)*y(n*l+k)
		enddo
	enddo
	do k=1,n
		do l=1,j-1
			y(n*k+j)=y(n*k+j)-gsc(l)*y(n*k+l)
		enddo
	enddo
	znorm(j)=0.d0
	do k=1,n
		znorm(j)=znorm(j)+y(n*k+j)**2
	enddo
	znorm(j)=dsqrt(znorm(j))
	do k=1,n
		y(n*k+j)=y(n*k+j)/znorm(j)
	enddo
enddo

return
end subroutine Gram_Schmidt_orthogonormalization

!**************************************************************

subroutine Lyap_exponents(n,t,suma,sumkt,znorm)
implicit none
integer,intent(in)		:: n
real(kind=8),intent(inout)	:: t, suma(n), sumkt(n), znorm(n)
integer				:: i, k

do k=1,n
	suma(k)=suma(k)+dlog(znorm(k))
enddo
do k=1,n
	sumkt(k)=suma(k)/t			
enddo

do i=1,n
	do k=1,n-1
		if (sumkt(k)<sumkt(k+1)) then
			call swap_Lyap_exp(n,sumkt,k,k+1)
		endif
	enddo
enddo

return
end subroutine Lyap_exponents

!**************************************************************

subroutine write_Lyap_exponents(n,sumkt,t,numfile_info)
implicit none
integer,intent(in)	:: n, numfile_info
real(kind=8),intent(in)	:: sumkt(n), t
integer			:: i, k

do i=1,n
	do k=1,n-1
		if (sumkt(k)<sumkt(k+1)) then
			call swap_Lyap_exp(n,sumkt,k,k+1)
		endif
	enddo
enddo
do k=1,n
	write(numfile_info,'(1x,"Lyapunov exponent(",i3,") = ",f26.16," after time ",es13.6)') k,sumkt(k),t
	call flush(numfile_info)
enddo

return
end subroutine write_Lyap_exponents

!**************************************************************

subroutine sum_Lyap_exponents(n,sumkt,sum_Lyap_exp)
implicit none
integer,intent(in)		:: n
real(kind=8),intent(in)		:: sumkt(n)
real(kind=8),intent(out)	:: sum_Lyap_exp
integer				:: i

sum_Lyap_exp=0.d0
do i=1,n
	sum_Lyap_exp=sum_Lyap_exp+sumkt(i)
enddo

return
end subroutine sum_Lyap_exponents

!**************************************************************

subroutine Lyapunov_Numbers(n,sumkt,Lyap_number)
implicit none
integer,intent(in)		:: n
real(kind=8),intent(out)	:: Lyap_number(n)
real(kind=8),intent(in)		:: sumkt(n)
integer				:: i

do i=1,n
	Lyap_number(i)=dexp(sumkt(i))
enddo

return
end subroutine Lyapunov_Numbers

!**************************************************************

subroutine write_Lyap_numbers(n,Lyap_number,t,numfile_info)
implicit none
integer,intent(in)	:: n, numfile_info
real(kind=8),intent(in)	:: Lyap_number(n), t
integer			:: k

write(numfile_info,*)
call flush(numfile_info)
do k=1,n
	write(numfile_info,'(1x,"Lyapunov number(",i3,") = ",f26.16," after time ",es13.6)') k,Lyap_number(k),t
	call flush(numfile_info)
enddo

return
end subroutine write_Lyap_numbers

!**************************************************************

subroutine Lyapunov_Dimension(n,h,Lyap_dimension)
implicit none
integer,intent(in)		:: n
real(kind=8),intent(inout)	:: h(n)
real(kind=8),intent(out)	:: Lyap_dimension
real(kind=8)			:: partial_sum, final_sum
integer				:: i, p

call my_sort_bubble(n,h)
partial_sum=0.d0
do i=1,n
	partial_sum=partial_sum+h(i)
	if (partial_sum>=0.d0) then
		final_sum=partial_sum
		p=i
	endif
enddo
if (p==0) then
	Lyap_dimension=0.d0
	return
endif
if ((p<n).and.(final_sum>=0.d0)) then
	Lyap_dimension=p+(1.d0/dabs(h(p+1)))*final_sum
	return
endif
if (p==n) then
	Lyap_dimension=n
	return
endif

return
end subroutine Lyapunov_Dimension

!**************************************************************

subroutine swap(a,b)
implicit none
real(kind=8),intent(inout)	:: a, b
real(kind=8)			:: temp

temp=a
a=b
b=temp

return
end subroutine swap

!*********************************************************

subroutine my_sort_bubble(n,h)
implicit none
integer,intent(in)		:: n
real(kind=8),intent(inout)	:: h(n)
integer				:: changes
integer				:: i

changes=1
do while (changes==1)
	changes=0
	do i=1,n-1
		if (h(i+1)>h(i)) then
			call swap(h(i+1),h(i))			
			changes=1
		endif
	enddo
enddo

return
end subroutine my_sort_bubble

!**************************************************************

subroutine write_Lyap_Dimension(Lyap_Dimension,t,numfile_info)
implicit none
integer,intent(in)	:: numfile_info
real(kind=8),intent(in)	:: Lyap_dimension, t

write(numfile_info,*)
call flush(numfile_info)
write(numfile_info,'(1x,"Lyapunov dimension = ",f26.16," after time ",es13.6)') Lyap_dimension,t
call flush(numfile_info)

return
end subroutine write_Lyap_Dimension

!**************************************************************

subroutine classification_of_the_type_of_attractor(n,sumkt,tol_chaos,numfile_info)
implicit none
integer,intent(in)	:: n, numfile_info
real(kind=8),intent(in)	:: sumkt(n+1), tol_chaos
integer			:: i, j, index

call my_sort_bubble(n,sumkt)
i=1
do while ((sumkt(i)<0.d0).and.(i<=n))
	i=i+1
enddo
if (i-1==n) then
	write(numfile_info,*)
	call flush(numfile_info)
	write(numfile_info,'(1x,"Attractor ===> Possible FIXED POINT with tolerance :",es13.6)') tol_chaos
	call flush(numfile_info)
endif

if (dabs(sumkt(1))<tol_chaos) then
	i=2
	do while ((sumkt(i)<0.d0).and.(i<=n))
		i=i+1
	enddo
	if (i-1==n) then
		write(numfile_info,*)
		call flush(numfile_info)
		write(numfile_info,'(1x,"Attractor ===> Possible LIMIT CYCLE with tolerance :",es13.6)') tol_chaos
		call flush(numfile_info)
	endif
endif

j=1
do while ((dabs(sumkt(j))<tol_chaos).and.(j<=n))
	j=j+1
enddo
i=j
index=j
do while ((sumkt(i)<0.d0).and.(j>2).and.(i<=n))
	i=i+1
enddo
if ((i-1==n).and.(j>2).and.(j<=n)) then
	write(numfile_info,*)
	call flush(numfile_info)
	write(numfile_info,'(1x,"Attractor ===> Possible T^",i3," torus with tolerance :",es13.6)') index-1,tol_chaos
	call flush(numfile_info)
endif

i=1
do while ((dabs(sumkt(i))<tol_chaos).and.(i<=n))
	i=i+1
enddo
if (i-1==n) then
	index=i
	write(numfile_info,*)
	call flush(numfile_info)
	write(numfile_info,'(1x,"Attractor ===> Possible T^",i3," torus with tolerance :",es13.6)') index-1,tol_chaos
	call flush(numfile_info)
endif

i=1
do while ((sumkt(i)>0.d0).and.(i<=n))
	i=i+1
enddo
if (dabs(sumkt(i))<tol_chaos) then
	do while ((sumkt(i+1)<0.d0).and.(i<=n))
		i=i+1
	enddo
	if (i-1==n) then
		write(numfile_info,*)
		call flush(numfile_info)
		write(numfile_info,'(1x,"Attractor ===> Possible STRANGE or CHAOTIC ATTRACTOR with tolerance :",es13.6)') tol_chaos
		call flush(numfile_info)
	endif
endif

return
end subroutine classification_of_the_type_of_attractor

!**************************************************************

subroutine chaos_or_integrability(n,sumkt,tol_chaos,numfile_info)
implicit none
integer,intent(in)	:: n, numfile_info
real(kind=8),intent(in)	:: sumkt(n), tol_chaos
real(kind=8)		:: maximum_Lyap_exp
integer			:: i

i=1
do while ((dabs(sumkt(i))<tol_chaos).and.(i<=n))
	i=i+1
enddo
if (i-1==n) then
	write(numfile_info,*)
	call flush(numfile_info)
	write(numfile_info,'(1x,"Possible Inference ===> INTEGRABLE CASE with tolerance :",es13.6)') tol_chaos
	call flush(numfile_info)
	return
endif

maximum_Lyap_exp=sumkt(1)
do i=1,n
	maximum_Lyap_exp=dmax1(maximum_Lyap_exp,sumkt(i))
enddo
if (maximum_Lyap_exp>0.d0) then
	write(numfile_info,*)
	call flush(numfile_info)
	write(numfile_info,'(1x,"Possible Inference ===> EXISTENCE OF CHAOS")')
	call flush(numfile_info)
endif

return
end subroutine chaos_or_integrability

!**************************************************************

subroutine write_initial_data(n,NN,NNN,initial_condition,t0,tn,acc,nrg,choosemode,numfile_info)
implicit none
integer,intent(in)	:: n, NN, NNN, choosemode, numfile_info
real(kind=8),intent(in)	:: initial_condition(n), t0, tn, acc, nrg
real(kind=8),external	:: second_integral
real(kind=8)		:: gamma, epsilon
integer			:: i
common /param1/ gamma, epsilon

write(numfile_info,*)
call flush(numfile_info)
write(numfile_info,'(1x,"****************************************************")')
call flush(numfile_info)
write(numfile_info,'(13x,"BEC Lyapunov Exponents")')
call flush(numfile_info)
write(numfile_info,'(1x,"****************************************************")')
call flush(numfile_info)
write(numfile_info,*)
call flush(numfile_info)
if (choosemode==1) then
	write(numfile_info,'(1x,"IN PHASE MODE")')
	call flush(numfile_info)
endif
if (choosemode==2) then
	write(numfile_info,'(1x,"OUT OF PHASE MODE")')
	call flush(numfile_info)
endif
write(numfile_info,*)
call flush(numfile_info)
write(numfile_info,'(1x,"Total number of particles=",3x,i3,1x,"particles")') NN
call flush(numfile_info)
write(numfile_info,*)
call flush(numfile_info)
write(numfile_info,'(1x,"Initial Condition = ( ",<n>f26.16,")")') (initial_condition(i),i=1,n)
call flush(numfile_info)
write(numfile_info,*)
call flush(numfile_info)
write(numfile_info,'(1x,"Time interval = [ ",f26.16,", ",es13.6," ]")') t0,tn
call flush(numfile_info)
write(numfile_info,*)
call flush(numfile_info)
write(numfile_info,'(1x,"GAMMA => ",f26.16)') gamma
call flush(numfile_info)
write(numfile_info,*)
call flush(numfile_info)
write(numfile_info,'(1x,"EPSILON => ",f26.16)') epsilon
call flush(numfile_info)
write(numfile_info,*)
call flush(numfile_info)
write(numfile_info,'(1x,"Total ENERGY of the BEC Hamiltonian => ",1x,f26.16)') nrg
call flush(numfile_info)
write(numfile_info,*)
call flush(numfile_info)
write(numfile_info,'(1x,"Energy per perticle=",1x,f26.16)') nrg/NN
call flush(numfile_info)
write(numfile_info,*)
call flush(numfile_info)
write(numfile_info,'(1x,"Second integral value => ",1x,f26.16)') second_integral(initial_condition,n)
call flush(numfile_info)
write(numfile_info,*)
call flush(numfile_info)
write(numfile_info,'(1x,"Conservation accuracy of ENERGY: ",es13.6)') acc
call flush(numfile_info)
write(numfile_info,*)
call flush(numfile_info)

return
end subroutine write_initial_data

!**************************************************************

subroutine write_sum_Lyap_exponents(sum_Lyap_exp,t,numfile_info)
implicit none
integer,intent(in)	:: numfile_info
real(kind=8),intent(in)	:: sum_Lyap_exp, t

write(numfile_info,*)
call flush(numfile_info)
write(numfile_info,'(1x,"Sum of Lyapunov exponents = ",f26.16," after time ",es13.6)') sum_Lyap_exp,t
call flush(numfile_info)

return
end subroutine write_sum_Lyap_exponents

!**************************************************************

subroutine swap_Lyap_exp(n,sumkt,index_1,index_2)
implicit none
integer,intent(in)		:: n, index_1, index_2
real(kind=8),intent(inout)	:: sumkt(n)
real(kind=8)			:: temp

temp=sumkt(index_1)
sumkt(index_1)=sumkt(index_2)
sumkt(index_2)=temp

return
end subroutine swap_Lyap_exp

!**************************************************************

subroutine write_my_name(numfile_info)
implicit none
integer,intent(in)	:: numfile_info

write(numfile_info,*)
call flush(numfile_info)
write(numfile_info,'(1x,"Programmed by Chris Antonopoulos")')
call flush(numfile_info)
close(numfile_info)

return
end subroutine write_my_name

!**************************************************************

subroutine problem_accuracy(t,hamiltonianenergy,numfile_info)
implicit none
integer,intent(in)	:: numfile_info
real(kind=8),intent(in)	:: t, hamiltonianenergy

write(numfile_info,'(1x,"CANNOT MAINTAIN CONSERVATION ACCURACY of ENERGY at time ",es13.6,1x,"for the energy=",1x,f26.16)') t,hamiltonianenergy
call flush(numfile_info)
write(numfile_info,*)
call flush(numfile_info)
write(numfile_info,'(1x,"Programmed by Chris Antonopoulos")')
call flush(numfile_info)
write(*,'(1x,"CANNOT MAINTAIN CONSERVATION ACCURACY of ENERGY at time ",es13.6,1x,"for the energy=",1x,f26.16)') t,hamiltonianenergy

return
end subroutine problem_accuracy

!**************************************************************

subroutine write_used_time(time_begin,numfile_info)
!use dfport
implicit none
integer,intent(in)	:: numfile_info
real(kind=4),intent(in)	:: time_begin
real(kind=4)		:: time_end
!real(kind=4)		:: time_used, ta(2)

!time_used=etime(ta)
!write(numfile_info,*)
!call flush(numfile_info)
!write(numfile_info,'(1x,"Program has used ",f10.3," seconds of CPU time.")') time_used
!call flush(numfile_info)
!write(numfile_info,'(1x,"This includes ",f10.3," seconds of USER time and ",f10.3," seconds of SYSTEM time.")') ta(1),ta(2)
!call flush(numfile_info)
write(numfile_info,*)
call flush(numfile_info)

call cpu_time(time_end)
write(numfile_info,'(1x,"Time of operation was",1x,f10.3,1x,"seconds")') time_end-time_begin
call flush(numfile_info)
!write(numfile_info,*)
!call flush(numfile_info)
!write(numfile_info,'(1x,"Programmed by Chris Antonopoulos")')
!call flush(numfile_info)

!write(*,*)
!write(*,'(1x,"Program has used ",f10.3," seconds of CPU time.")') time_used
!write(*,'(1x,"This includes ",f10.3," seconds of USER time and ",f10.3," seconds of SYSTEM time.")') ta(1),ta(2)
!write(*,*)

write(*,'(1x,"Time of operation was",1x,f10.3,1x,"seconds")') time_end-time_begin
write(*,*)
write(*,'(1x,"Programmed by Chris Antonopoulos")')

return
end subroutine write_used_time

!**************************************************************

subroutine read_input_parameters(perturbation,t0,tf,ndof,numfile_info,numfile_Lyap_exps,filename_input_parameters,numfile_LEs_plot_gnu,numfile_IcLE_and_maxdK_dt_vs_nrg,numfile_traj,gamma,eps,h1,hmin,acc,tol_chaos,choosemode)
implicit none
real(kind=8),intent(out)	:: perturbation, t0, tf, gamma, eps, h1, hmin, acc, tol_chaos
integer(kind=4),intent(out)	:: ndof, numfile_info, numfile_Lyap_exps, numfile_LEs_plot_gnu, numfile_IcLE_and_maxdK_dt_vs_nrg, numfile_traj, choosemode
character(*),intent(in)		:: filename_input_parameters

open(1,file=filename_input_parameters,status='old')
read(1,*)
read(1,*) perturbation
read(1,*)
read(1,*)
read(1,*) t0
read(1,*)
read(1,*)
read(1,*) tf
read(1,*)
read(1,*)
read(1,*) ndof
read(1,*)
read(1,*)
read(1,*) eps
read(1,*)
read(1,*)
read(1,*) h1
read(1,*)
read(1,*)
read(1,*) hmin
read(1,*)
read(1,*)
read(1,*) acc
read(1,*)
read(1,*)
read(1,*) tol_chaos
read(1,*)
read(1,*)
read(1,*) choosemode
read(1,*)
read(1,*)
read(1,*) gamma
close(1)

numfile_info=1
numfile_Lyap_exps=2
numfile_LEs_plot_gnu=3
numfile_IcLE_and_maxdK_dt_vs_nrg=4
numfile_traj=5

return
end subroutine read_input_parameters

!**************************************************************

subroutine LEs_plot_gnu_file(numfile_LEs_plot_gnu,char_NN,char_nrg)
implicit none
integer(kind=4),intent(in)	:: numfile_LEs_plot_gnu
character(*),intent(in)		:: char_NN, char_nrg

open(numfile_LEs_plot_gnu,file='LEs_plot_'//char_NN//'_nrg='//char_nrg//'.gnu',status='replace')
write(numfile_LEs_plot_gnu,'("set output ''LEs_vs_t_",a,"_nrg=",a,".eps''")') char_NN, char_nrg
write(numfile_LEs_plot_gnu,'("set term postscript color enhanced")')
write(numfile_LEs_plot_gnu,'("set size 0.6,0.6")')

write(numfile_LEs_plot_gnu,'("set format x ""10^{%L}""")')
write(numfile_LEs_plot_gnu,'("set format y ""10^{%L}""")')

write(numfile_LEs_plot_gnu,'("set logscale xy")')
write(numfile_LEs_plot_gnu,'("set xlabel ''t''")')
write(numfile_LEs_plot_gnu,'("#set ylabel ''LEs''")')

write(numfile_LEs_plot_gnu,'("#set title ""Lyapunov exponents"" font ""Times-Roman,20""")')

write(numfile_LEs_plot_gnu,'("#unset key")')
write(numfile_LEs_plot_gnu,'("#set label 1 ''(a)'' at 3.5e7, 1.1 center")')

write(numfile_LEs_plot_gnu,'("plot [1e1:][] ''BEC_LE_data_",a,"_nrg=",a,".txt'' u 1:2 w l title ""{/Symbol l}_1"", ''BEC_LE_data_",a,"_nrg=",a,".txt'' u 1:3 w l title ""{/Symbol l}_2"", ''BEC_LE_data_",a,"_nrg=",a,".txt'' u 1:4 w l title ""{/Symbol l}_3"", 1/x w l title ""1/t""")') char_NN, char_nrg, char_NN, char_nrg, char_NN, char_nrg
close(numfile_LEs_plot_gnu)

return
end subroutine LEs_plot_gnu_file
!**************************************************************
