SUBROUTINE odeint(ystart,nvar,x1,x2,eps,h1,hmin,nok,nbad,derivs,rkqs,acc,nn2,znorm,gsc,suma,sumkt,ndof,flag,hamiltoniannrg,x,numfile_traj,sum_time_average_kinetic_energy_y1,numfile_Lyap_exps,numfile_info)
real(kind=8),external		:: hamiltonian_function, solved_hamiltonian_function, kinetic_energy_y1
INTEGER nbad,nok,nvar,KMAXX,MAXSTP,NMAX,nn2,ndof,flag,numfile_traj,numfile_Lyap_exps,numfile_info
REAL*8 eps,h1,hmin,x1,x2,ystart(nvar),TINY,acc,znorm(nn2),gsc(nn2),suma(nn2),sumkt(nn2),Lyap_number(nn2),hamiltoniannrg,sum_time_average_kinetic_energy_y1,kinetic_energy_y1_old
EXTERNAL derivs,rkqs
!PARAMETER (MAXSTP=100000000000000,NMAX=48000,KMAXX=644,TINY=1.e-30)
PARAMETER (MAXSTP=1000000000,NMAX=60000,KMAXX=8,TINY=1.e-30)
!Runge-Kutta driver with adaptive stepsize control. Integrate the starting values ystart(1:nvar)
!from x1 to x2 with accuracy eps, storing intermediate results in the common block /path/.
!h1 should be set as a guessed first stepsize, hmin as the minimum allowed stepsize (can
!be zero). On output nok and nbad are the number of good and bad (but retried and
!fixed) steps taken, and ystart is replaced by values at the end of the integration interval.
!derivs is the user-supplied subroutine for calculating the right-hand side derivative, while
!rkqs is the name of the stepper routine to be used. /path/ contains its own information
!about how often an intermediate value is to be stored.
INTEGER i,kmax,kount,nstp
REAL*8 dxsav,h,hdid,hnext,x,xsav,dydx(NMAX),xp(KMAXX),y(NMAX),yp(NMAX,KMAXX),yscal(NMAX)
COMMON /path/ kmax,kount,dxsav,xp,yp

flag=0

multiple=0

!User storage for intermediate results. Preset dxsav and kmax.

!dxsav=h1	;	kmax=KMAXX

!If kmax=0 no intermediate savings are done
dxsav=0.d0	;	kmax=0

x=x1
h=dsign(h1,x2-x1)
nok=0
nbad=0
kount=0

do 11 i=1,nvar
	y(i)=ystart(i)
11 enddo

if (kmax.gt.0) xsav=x-2.*dxsav !Assures storage of first step.

do 16 nstp=1,MAXSTP !Take at most MAXSTP steps.
	call derivs(x,y,dydx)

	do 12 i=1,nvar
		!Scaling used to monitor accuracy. This general-purpose choice can be modified if need be.
		yscal(i)=dabs(y(i))+dabs(h*dydx(i))+TINY
	12 enddo

	if(kmax.gt.0)then
		if(dabs(x-xsav).gt.dabs(dxsav)) then !Store intermediate results.
			if(kount.lt.kmax-1)then
				kount=kount+1
				xp(kount)=x
				do 13 i=1,nvar
					yp(i,kount)=y(i)
				13 enddo
				xsav=x
			endif
		endif
	endif

	if((x+h-x2)*(x+h-x1).gt.0.) h=x2-x !If stepsize can overshoot, decrease.

	kinetic_energy_y1_old=kinetic_energy_y1(y,nvar)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	call rkqs(y,dydx,nvar,x,h,eps,yscal,hdid,hnext,derivs)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

	if (dabs(hamiltonian_function(y,nvar))>=acc) then
		call problem_accuracy(x,hamiltoniannrg,numfile_info)
		flag=1
		return
	endif

	sum_time_average_kinetic_energy_y1=sum_time_average_kinetic_energy_y1+dabs(kinetic_energy_y1(y,nvar)-kinetic_energy_y1_old)
	
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	! here we write the trajectory for Ezequiel to compute the MIR
	if (dmod(x,1.d0)<hdid) then
		write(numfile_traj,'(es13.6,2f26.16)') x, y(1), y(ndof)
		call flush(numfile_traj)
	endif
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

	call Gram_Schmidt_orthogonormalization(nn2,nvar,znorm,y)
	call Lyap_exponents(nn2,x,suma,sumkt,znorm)
	if (dabs(x-10.d0**(1.d0+multiple*.05d0))<hdid) then
!	if (dmod(x,10.d0)<hdid) then
		write(numfile_Lyap_exps,'(es13.6,1x,<nn2>f26.16)') x, (sumkt(i),i=1,nn2)
		call flush(numfile_Lyap_exps)
		multiple=multiple+1
	endif

	if(hdid.eq.h)then
		nok=nok+1
	else
		nbad=nbad+1
	endif
	if((x-x2)*(x2-x1).ge.0.)then !Are we done?
		do 14 i=1,nvar
			ystart(i)=y(i)
		14 enddo
		if(kmax.ne.0)then
			kount=kount+1 !Save final step.
			xp(kount)=x
			do 15 i=1,nvar
				yp(i,kount)=y(i)
			15 enddo
		endif
		return !Normal exit.
	endif
	if(dabs(hnext).lt.hmin) write(*,'("stepsize smaller than minimum in odeint at time=",es13.6)') x !pause 'stepsize smaller than minimum in odeint'
	h=hnext
16 enddo

write(*,'("too many steps in odeint at time=",es13.6)') x !pause 'too many steps in odeint'
return

END

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

SUBROUTINE bsstep(y,dydx,nv,x,htry,eps,yscal,hdid,hnext,derivs)
INTEGER nv,NMAX,KMAXX,IMAX
REAL*8 eps,hdid,hnext,htry,x,dydx(nv),y(nv),yscal(nv),SAFE1,SAFE2,REDMAX,REDMIN,TINY,SCALMX
PARAMETER (NMAX=60000,KMAXX=8,IMAX=KMAXX+1,SAFE1=.25,SAFE2=.7,REDMAX=1.e-5,REDMIN=.7,TINY=1.e-30,SCALMX=.1)
! C USES derivs,mmid,pzextr
!Bulirsch-Stoer step with monitoring of local truncation error to ensure accuracy and adjust
!stepsize. Input are the dependent variable vector y(1:nv) and its derivative dydx(1:nv)
!at the starting value of the independent variable x. Also input are the stepsize to be attempted
!htry, the required accuracy eps, and the vector yscal(1:nv) against which the
!error is scaled. On output, y and x are replaced by their new values, hdid is the stepsize
!that was actually accomplished, and hnext is the estimated next stepsize. derivs is the
!user-supplied subroutine that computes the right-hand side derivatives. Be sure to set htry
!on successive steps to the value of hnext returned from the previous step, as is the case
!if the routine is called by odeint.
!Parameters: NMAX is the maximum value of nv; KMAXX is the maximum row number used
!in the extrapolation; IMAX is the next row number; SAFE1 and SAFE2 are safety factors;
!REDMAX is the maximum factor used when a stepsize is reduced, REDMIN the minimum;
!TINY prevents division by zero; 1/SCALMX is the maximum factor by which a stepsize can
!be increased.
INTEGER i,iq,k,kk,km,kmax,kopt,nseq(IMAX)
REAL*8 eps1,epsold,errmax,fact,h,red,scale,work,wrkmin,xest,xnew,a(IMAX),alf(KMAXX,KMAXX),err(KMAXX),yerr(NMAX),ysav(NMAX),yseq(NMAX)
LOGICAL first,reduct
SAVE a,alf,epsold,first,kmax,kopt,nseq,xnew
EXTERNAL derivs
DATA first/.true./,epsold/-1./
DATA nseq /2,4,6,8,10,12,14,16,18/
if(eps.ne.epsold)then !A new tolerance, so reinitialize.
hnext=-1.e29 !Impossible values.
xnew=-1.e29
eps1=SAFE1*eps
a(1)=nseq(1)+1 !Compute work coefficients Ak.
do 11 k=1,KMAXX
a(k+1)=a(k)+nseq(k+1)
11 enddo
do 13 iq=2,KMAXX !Compute (k, q).
do 12 k=1,iq-1
alf(k,iq)=eps1**((a(k+1)-a(iq+1))/((a(iq+1)-a(1)+1.)*(2*k+1)))
12 enddo
13 enddo
epsold=eps
do 14 kopt=2,KMAXX-1 !Determine optimal row number for convergence.
if(a(kopt+1).gt.a(kopt)*alf(kopt-1,kopt))goto 1
14 enddo
1 kmax=kopt
endif
h=htry
do 15 i=1,nv !Save the starting values.
ysav(i)=y(i)
15 enddo
if(h.ne.hnext.or.x.ne.xnew)then !A new stepsize or a new integration: re-establish
!the order window.
first=.true.
kopt=kmax
endif
reduct=.false.
2 do 17 k=1,kmax !Evaluate the sequence of modified midpoint
!integrations.
xnew=x+h
if(xnew.eq.x)write(*,'("step size underflow in bsstep at time=",es13.6)') x !pause 'step size underflow in bsstep'
call mmid(ysav,dydx,nv,x,h,nseq(k),yseq,derivs)
xest=(h/nseq(k))**2 !Squared, since error series is even.
call pzextr(k,xest,yseq,y,yerr,nv) !Perform extrapolation.
if(k.ne.1)then !Compute normalized error estimate .(k).
errmax=TINY
do 16 i=1,nv
errmax=dmax1(errmax,dabs(yerr(i)/yscal(i)))
16 enddo
errmax=errmax/eps !Scale error relative to tolerance.
km=k-1
err(km)=(errmax/SAFE1)**(1./(2*km+1))
endif
if(k.ne.1.and.(k.ge.kopt-1.or.first))then !In order window.
if(errmax.lt.1.)goto 4 !Converged.
if(k.eq.kmax.or.k.eq.kopt+1)then !Check for possible stepsize reduction.
red=SAFE2/err(km)
goto 3
else if(k.eq.kopt)then
if(alf(kopt-1,kopt).lt.err(km))then
red=1./err(km)
goto 3
endif
else if(kopt.eq.kmax)then
if(alf(km,kmax-1).lt.err(km))then
red=alf(km,kmax-1)*SAFE2/err(km)
goto 3
endif
else if(alf(km,kopt).lt.err(km))then
red=alf(km,kopt-1)/err(km)
goto 3
endif
endif
17 enddo
3 red=dmin1(red,REDMIN) !Reduce stepsize by at least REDMIN and at most REDMAX. 
red=dmax1(red,REDMAX)
h=h*red
reduct=.true.
goto 2 !Try again.
4 x=xnew !Successful step taken.
hdid=h
first=.false.
wrkmin=1.e35 !Compute optimal row for convergence and corresponding stepsize. 
do 18 kk=1,km
fact=dmax1(err(kk),SCALMX)
work=fact*a(kk+1)
if(work.lt.wrkmin)then
scale=fact
wrkmin=work
kopt=kk+1
endif
18 enddo
hnext=h/scale
if(kopt.ge.k.and.kopt.ne.kmax.and..not.reduct)then !Check for possible order increase, but not if stepsize was just reduced.
fact=dmax1(scale/alf(kopt-1,kopt),SCALMX)
if(a(kopt+1)*fact.le.wrkmin)then
hnext=h/fact
kopt=kopt+1
endif
endif
return
END

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!The polynomial extrapolation routine is based on the same algorithm as polint
!3.1. It is simpler in that it is always extrapolating to zero, rather than to an arbitrary
!value. However, it is more complicated in that it must individually extrapolate each
!component of a vector of quantities.
SUBROUTINE pzextr(iest,xest,yest,yz,dy,nv)
INTEGER iest,nv,IMAX,NMAX
REAL*8 xest,dy(nv),yest(nv),yz(nv)
PARAMETER (IMAX=13,NMAX=60000)
!Use polynomial extrapolation to evaluate nv functions at x = 0 by fitting a polynomial to a
!sequence of estimates with progressively smaller values x = xest, and corresponding function
!vectors yest(1:nv). This call is number iest in the sequence of calls. Extrapolated
!function values are output as yz(1:nv), and their estimated error is output as dy(1:nv).
!Parameters: Maximum expected value of iest is IMAX; of nv is NMAX.
INTEGER j,k1
REAL*8 delta,f1,f2,q,d(NMAX),qcol(NMAX,IMAX),x(IMAX)
SAVE qcol,x
x(iest)=xest !Save current independent variable.
do 11 j=1,nv
dy(j)=yest(j)
yz(j)=yest(j)
11 enddo
if(iest.eq.1) then !Store first estimate in first column.
do 12 j=1,nv
qcol(j,1)=yest(j)
12 enddo
else
do 13 j=1,nv
d(j)=yest(j)
13 enddo
do 15 k1=1,iest-1
delta=1./(x(iest-k1)-xest)
f1=xest*delta
f2=x(iest-k1)*delta
do 14 j=1,nv !Propagate tableau 1 diagonal more.
q=qcol(j,k1)
qcol(j,k1)=dy(j)
delta=d(j)-q
dy(j)=f1*delta
d(j)=f2*delta
yz(j)=yz(j)+dy(j)
14 enddo
15 enddo
do 16 j=1,nv
qcol(j,iest)=dy(j)
16 enddo
endif
return
END

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

SUBROUTINE rzextr(iest,xest,yest,yz,dy,nv)
INTEGER iest,nv,IMAX,NMAX
REAL*8 xest,dy(nv),yest(nv),yz(nv)
PARAMETER (IMAX=13,NMAX=60000)
!Exact substitute for pzextr, but uses diagonal rational function extrapolation instead of
!polynomial extrapolation.
INTEGER j,k
REAL*8 b,b1,c,ddy,v,yy,d(NMAX,IMAX),fx(IMAX),x(IMAX)
SAVE d,x
x(iest)=xest !Save current independent variable.
if(iest.eq.1) then
do 11 j=1,nv
yz(j)=yest(j)
d(j,1)=yest(j)
dy(j)=yest(j)
11 enddo
else
do 12 k=1,iest-1
fx(k+1)=x(iest-k)/xest
12 enddo
do 14 j=1,nv !Evaluate next diagonal in tableau.
yy=yest(j)
v=d(j,1)
c=yy
d(j,1)=yy
do 13 k=2,iest
b1=fx(k)*v
b=b1-c
if(b.ne.0.) then
b=(c-v)/b
ddy=c*b
c=b1*b
else !Care needed to avoid division by 0.
ddy=v
endif
if (k.ne.iest) v=d(j,k)
d(j,k)=ddy
yy=yy+ddy
13 enddo
dy(j)=ddy
yz(j)=yy
14 enddo
endif
return
END

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


SUBROUTINE mmid(y,dydx,nvar,xs,htot,nstep,yout,derivs)
INTEGER nstep,nvar,NMAX
REAL*8 htot,xs,dydx(nvar),y(nvar),yout(nvar)
EXTERNAL derivs
PARAMETER (NMAX=60000)
!Modified midpoint step. Dependent variable vector y(1:nvar) and its derivative vector
!dydx(1:nvar) are input at xs. Also input is htot, the total step to be made, and nstep,
!the number of substeps to be used. The output is returned as yout(1:nvar), which need
!not be a distinct array from y; if it is distinct, however, then y and dydx are returned
!undamaged.
INTEGER i,n
REAL*8 h,h2,swap,x,ym(NMAX),yn(NMAX)

h=htot/nstep !Stepsize this trip.

do 11 i=1,nvar
	ym(i)=y(i)
	yn(i)=y(i)+h*dydx(i) !First step.
11 enddo

x=xs+h

call derivs(x,yn,yout) !Will use yout for temporary storage of derivatives.

h2=2.*h
do 13 n=2,nstep !General step.
	do 12 i=1,nvar
		swap=ym(i)+h2*yout(i)
		ym(i)=yn(i)
		yn(i)=swap
	12 enddo
	x=x+h
	call derivs(x,yn,yout)
13 enddo

do 14 i=1,nvar !Last step.
	yout(i)=0.5*(ym(i)+yn(i)+h*yout(i))
14 enddo

return
END

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
